/*
  CPP_LANG=C+OPENMP+MPI
  CPP_NUM_CORES = 32
  CPP_PROCESSES_PER_NODE 8
  CPP_PROBLEM=E
 */

/* RECORD
Andrés Sánchez Pascual
student of Algorithms and Parallel programming
at the Unviersity of Murcia
June 26, 2012
time 1721 msec
speedup 12.29*/

#include <stdlib.h>
#include <stdio.h>
#include <mpi.h>
#include <math.h>

struct Nodo
{
    int *solucion;
    int *pesopormochila;
};

void generarArbol(int nivel, int niveles, int mochilas, int objetos, int *pesos, int *capacidad, int *solucion, int *pesopormochila, struct Nodo *arbol, int *n)
{
    int i;

    if (nivel == niveles)
    {
        arbol[*n].solucion = (int*) calloc(objetos, sizeof(int));
        arbol[*n].pesopormochila = (int*) calloc(mochilas, sizeof(int));
        
        for (i = 0; i < objetos; i++)
        {
            arbol[*n].solucion[i] = solucion[i];
            //printf("%d ", solucion[i]);
        }
        
        //printf("\n");
        
        for (i = 0; i < mochilas; i++)
        {
            arbol[*n].pesopormochila[i] = pesopormochila[i];
            //printf("%d ", pesopormochila[i]);
        }

        //printf("\n");        

        *n = *n + 1;
    }
    else
    {
        for (i = 0; i < mochilas; i++)
        {
            solucion[nivel] = i;
            pesopormochila[solucion[nivel]] += pesos[nivel];
            
            if (pesopormochila[solucion[nivel]] <= capacidad[solucion[nivel]])
            {    
                generarArbol(nivel + 1, niveles, mochilas, objetos, pesos, capacidad, solucion, pesopormochila, arbol, n);                
            }
            
            pesopormochila[solucion[nivel]] -= pesos[nivel];
        }
    }
}

/**
 * Esta función permite enviar los datos de entrada del problema.
 * Es llamada únicamente por el proceso maestro.
 * 
 * @param mochilas Número de mochilas
 * @param objetos Número de objetos
 * @param capacidad Capacidades de las mochilas
 * @param peso Pesos de los objetos
 * @param afinidad Matriz de afinidades mochila / objeto
 */

void enviarEntrada(int mochilas, int objetos, int* capacidad, int* peso, int *afinidad)
{
    MPI_Bcast(&mochilas, 1, MPI_INT, 0, MPI_COMM_WORLD);
    MPI_Bcast(&objetos, 1, MPI_INT, 0, MPI_COMM_WORLD);
    MPI_Bcast(capacidad, mochilas, MPI_INT, 0, MPI_COMM_WORLD);
    MPI_Bcast(peso, objetos, MPI_INT, 0, MPI_COMM_WORLD);
    MPI_Bcast(afinidad, objetos*objetos, MPI_INT, 0, MPI_COMM_WORLD);
}

/**
 * Esta función permite recibir los datos de entrada del problema.
 * Es llamada por los procesos esclavos.
 * 
 * @param mochilas Número de mochilas
 * @param objetos Número de objetos
 * @param capacidad Capacidades de las mochilas
 * @param peso Pesos de los objetos
 * @param afinidad Matriz de afinidades mochila / objeto
 * 
 */

void recibirEntrada(int* mochilas, int* objetos, int** capacidad, int** peso, int **afinidad)
{
    MPI_Bcast(mochilas, 1, MPI_INT, 0, MPI_COMM_WORLD);
    MPI_Bcast(objetos, 1, MPI_INT, 0, MPI_COMM_WORLD);

    *capacidad = (int *) malloc(sizeof (int) *(*mochilas));
    *peso = (int *) malloc(sizeof (int) *(*objetos));
    *afinidad = (int *) malloc(sizeof (int) *(*objetos)*(*objetos));

    MPI_Bcast(*capacidad, (*mochilas), MPI_INT, 0, MPI_COMM_WORLD);
    MPI_Bcast(*peso, (*objetos), MPI_INT, 0, MPI_COMM_WORLD);
    MPI_Bcast(*afinidad, (*objetos)*(*objetos), MPI_INT, 0, MPI_COMM_WORLD);
}

/**
 * Envía un nodo a un proceso esclavo
 * @param n Nodo a mandar
 * @param proceso Proceso al que se manda
 * @param nivel Nivel generado
 * @param objetos Número de objetos
 * @param mochilas Número de mochilas
 */

void enviarNodo(struct Nodo* n, int proceso, int nivel, int objetos, int mochilas)
{
    int length = objetos + mochilas + 1;
    int size = sizeof(int) * length;
    int *buf = (int*) malloc(size);
    int p = 0;
    
    MPI_Pack(n -> solucion, objetos, MPI_INT, buf, size, &p, MPI_COMM_WORLD);
    MPI_Pack(n -> pesopormochila, mochilas, MPI_INT, buf, size, &p, MPI_COMM_WORLD);
    MPI_Pack(&nivel, 1, MPI_INT, buf, size, &p, MPI_COMM_WORLD);
    MPI_Send(buf, length, MPI_INT, proceso, 0, MPI_COMM_WORLD);

    free(buf);
}

/**
 * Recibe un nodo del proceso maestro
 * @param nivel Nivel generado
 * @param solucion Solución parcial
 * @param pesopormochila Pesos de la mochila
 * @param objetos Número de objetos
 * @param mochilas Número de mochilas
 */

void recibirNodo(int *nivel, int *solucion, int *pesopormochila, int objetos, int mochilas)
{
    int length = objetos + mochilas + 1;
    int size = sizeof(int) * length;
    int *buf = (int*) malloc(size);
    MPI_Status status;
    int p = 0;
    
    MPI_Recv(buf, length, MPI_INT, 0, MPI_ANY_TAG, MPI_COMM_WORLD, &status);
    MPI_Unpack(buf, size, &p, solucion, objetos, MPI_INT, MPI_COMM_WORLD);
    MPI_Unpack(buf, size, &p, pesopormochila, mochilas, MPI_INT, MPI_COMM_WORLD);
    MPI_Unpack(buf, size, &p, nivel, 1, MPI_INT, MPI_COMM_WORLD);

    free(buf);
}

/**
 * Genera el siguiente nivel del árbol.
 * 
 * @param nivel Nivel que se quiere generar
 * @param peso Vector de pesos
 * @param solucion Vector solución
 * @param pesopormochila Vector de pesos / mochila
 */

void generar(int nivel, int *peso, int *solucion, int *pesopormochila)
{
    if (solucion[nivel] != -1)
    {
        pesopormochila[solucion[nivel]] -= peso[nivel];
    }
    solucion[nivel]++;
    pesopormochila[solucion[nivel]] += peso[nivel];
}

/**
 * Comprueba si la configuración actual es solución del problema.
 * 
 * @param nivel Nivel actual
 * @param objetos Número de objetos
 * @param capacidad Vector capacidad de las mochilas
 * @param pesopormochila Vector de pesos / mochila
 * @param solucion Vector solución
 * @return Cierto / Falso
 */

int essolucion(int nivel, int objetos, int *capacidad, int *pesopormochila, int *solucion)
{
    return ((nivel == objetos - 1) && pesopormochila[solucion[nivel]] <= capacidad[solucion[nivel]]);
}

/**
 * Calcula el valor de la solución
 * 
 * @param objetos Número de objetos
 * @param solucion Vector solución
 * @param afinidad Vector de afinidades
 * @return Valor calculado de solución
 */

int calcularvalor(int objetos, int *solucion, int *afinidad)
{
    int valor = 0, i, j;

    for (i = 0; i < objetos; i++)
    {
        for (j = i + 1; j < objetos; j++)
        {
            if (solucion[i] == solucion[j])
            {
                valor += (afinidad[i * objetos + j] + afinidad[j * objetos + i]);
            }
        }
    }
    
    return valor;
}

/**
 * Comprueba si se da el criterio para seguir en esa rama.
 * 
 * @param nivel Nivel actual
 * @param objetos Número de objetos
 * @param capacidad Vector de capacidades
 * @param solucion Vector solución
 * @param pesopormochila Vector de pesos / mochila
 * @return Cierto / Falso
 */

int criterio(int nivel, int objetos, int *capacidad, int *solucion, int *pesopormochila)
{    
    return (nivel != objetos - 1 && pesopormochila[solucion[nivel]] <= capacidad[solucion[nivel]]);
}

/**
 * Comprueba en el nivel actual quedan más hermanos que explorar.
 * 
 * @param nivel Nivel actual
 * @param mochilas Número de mochilas
 * @param solucion Vector solución
 * @return  Cierto / Falso
 */

int mashermanos(int nivel, int mochilas, int *solucion)
{
    return (solucion[nivel] < mochilas - 1);
}

/**
 * Retrocede de nivel
 * 
 * @param nivel Nivel actual
 * @param mochilas Número de mochilas
 * @param peso Vector de pesos
 * @param solucion Vector solución
 * @param pesopormochila Vector de pesos / mochila
 */

void retroceder(int nivel, int mochilas, int *peso, int *solucion, int *pesopormochila)
{
    pesopormochila[mochilas - 1] -= peso[nivel];
    solucion[nivel] = -1;
}

/**
 * Parte donde se gestiona el paralelismo.
 * Punto de entrada al paralelismo MPI.
 * 
 * @param mochilas
 * @param objetos
 * @param capacidad
 * @param peso
 * @param afinidad
 * @param nodo
 * @param np
 * @return 
 */

int mochila(int mochilas, int objetos, int *capacidad, int *peso, int *afinidad, int nodo, int np)
{
    int i, n, ne = 1, enviados = 0, recibidos = 0, niveles = 2, valor, VOA = 0, total = 0;
    MPI_Status status;
    
    /*
     * Proceso maestro
     */
    if (nodo == 0)
    {
        // Envía la entrada a los procesos esclavos
        enviarEntrada(mochilas, objetos, capacidad, peso, afinidad);

        // Genera el árbol de soluciones parciales
        int niveles = round(log(np) / log(mochilas));
        int *solucion = (int*) calloc(objetos, sizeof(int));
        int *pesopormochila = (int*) calloc(mochilas, sizeof(int));
        struct Nodo *arbol = (struct Nodo*) malloc(sizeof(struct Nodo) * pow(mochilas, niveles));
        generarArbol(0, niveles, mochilas, objetos, peso, capacidad, solucion, pesopormochila, arbol, &total);

        // Envía todos los nodos
        for (i = 0; i < total; i ++)
        {
            //Envía el trozo
            enviarNodo(&arbol[i], ne, niveles-1, objetos, mochilas);

            //printf("Acabo de enviar %d a %d\n", i, ne);
            ne ++;
            enviados ++;
                    
            // Empieza a recibir trozos
            if (enviados >= np - 1)
            {
                MPI_Recv(&valor, 1, MPI_INT, MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &status);
                ne = status.MPI_SOURCE;
                recibidos ++;
                    
                // Es mejor?
                if (valor > VOA) VOA = valor;
                //printf("Valor: %d\nVOA: %d\n", valor, VOA);
            }
                
            //printf("recibidos: %d, enviados: %d, i: %d\n", recibidos, enviados, i);
        }

        // Acaba de recibir los trozos
        for (i = recibidos; i < enviados; i++)
        {
            MPI_Recv(&valor, 1, MPI_INT, MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &status);
            if (valor > VOA) VOA = valor;
            //printf("Valor: %d\nVOA: %d\n", valor, VOA);
        }
        
        // Finaliza todos los procesos esclavos
        n = -1;
        
        for (i = 1; i < np; i++)
        {
            enviarNodo(&arbol[0], i, n, objetos, mochilas);
            //printf("Acabo de enviar %d a %d\n", n, i);
        }
        
        // Libera la memoria creada localmente
        for (i = 0; i < total; i ++)
        {
            free(arbol[i].solucion);
            free(arbol[i].pesopormochila);
        }
        free(solucion);
        free(pesopormochila);
        free(arbol);
        
        return VOA;
    }
    /*
     * Proceso esclavo
     */
    else
    {
        int nivel, *solucion, *pesopormochila;
        
        //Recibe la entrada del proceso maestro
        recibirEntrada(&mochilas, &objetos, &capacidad, &peso, &afinidad);
        
        solucion = (int*) calloc(objetos, sizeof(int));
        pesopormochila = (int*) calloc(mochilas, sizeof(int));
        
        for (;;)
        {
            // Espera recibir un trozo
            recibirNodo(&nivel, solucion, pesopormochila, objetos, mochilas);

            //printf("Soy %d y he recibido: nivel %d\n", nodo, nivel);
            
            // Ya no quedan trozos, puede acabar
            if (nivel == -1) break;
           
            // Evalua su porción del problema        
            valor = mochilaAlg(mochilas, objetos, capacidad, peso, afinidad, solucion, pesopormochila, nivel);
            
            // Envía su solución
            MPI_Send(&valor, 1, MPI_INT, 0, 0, MPI_COMM_WORLD);
        }

        // Libera la memoria creada localmente
        free(solucion);
        free(pesopormochila);
    }
    
    return 1;
}

/**
 * Algoritmo de la mochila paralelo.
 * Recorre la parte que le toca del árbol.
 * 
 * @param mochilas
 * @param objetos
 * @param capacidad
 * @param peso
 * @param afinidad
 * @param nodo
 * @param np
 * @return 
 */

int mochilaAlg(int mochilas, int objetos, int *capacidad, int *peso, int *afinidad, int *solucion, int* pesopormochila, int nivelg)
{
    int valor = 0, VOA = 0, i, nivel;
    
    nivel = nivelg;
    
    for (i = nivel + 1; i < objetos; i++) solucion[i] = -1;
    
    //printf("Tras Inicializar\n");
    //escribir(nivel,mochilas,objetos,capacidad,peso,afinidad,solucion,pesopormochila);
    do
    {
        if (nivel != nivelg)
        {
            generar(nivel, peso, solucion, pesopormochila);
        }
        //printf("Tras Generar\n");
        //escribir(nivel,mochilas,objetos,capacidad,peso,afinidad,solucion,pesopormochila);
        if (essolucion(nivel, objetos, capacidad, pesopormochila, solucion) == 1)
        {
            valor = calcularvalor(objetos, solucion, afinidad);
            if (valor > VOA)
            {
                VOA = valor;
                //printf("Tras Essolucion, valor %d\n",valor);
                //escribir(nivel,mochilas,objetos,capacidad,peso,afinidad,solucion,pesopormochila);
            }
        }
        if (criterio(nivel, objetos, capacidad, solucion, pesopormochila) == 1)
        {
            //printf("Tras Criterio\n");
            //escribir(nivel,mochilas,objetos,capacidad,peso,afinidad,solucion,pesopormochila);
            nivel++;
        }
        else
        {
            while (nivel != nivelg && mashermanos(nivel, mochilas, solucion) == 0)
            {
                //printf("Tras Mashermanos\n");
                //escribir(nivel,mochilas,objetos,capacidad,peso,afinidad,solucion,pesopormochila);
                retroceder(nivel, mochilas, peso, solucion, pesopormochila);
                nivel--;
                //printf("Tras Retroceder\n");
                //escribir(nivel,mochilas,objetos,capacidad,peso,afinidad,solucion,pesopormochila);
            }
        }
    }
    while (nivel != nivelg);
    
    return VOA;
}



void escribir(int nivel,int mochilas,int objetos,int *capacidad,int *peso,int *afinidad,int *solucion,int *pesopormochila)
{
  int i,seguir;

  printf("En nivel %d\n",nivel);
  printf("  solucion:");
  for(i=0;i<objetos;i++)
    printf(" %d ",solucion[i]);
  printf("\n");
  printf("  pesos por mochila:");
  for(i=0;i<mochilas;i++)
    printf(" %d ",pesopormochila[i]);
  printf("\n");
  printf("\n");
}

